#!/usr/bin/perl -w

# annot8er_physprop.pl version 0.1
# calculates physical properties for peptide sequences 
# 
#
# Perl Modules - DBD:Pg and BioPerl
# DBD::Pg is part of the DBI is available from cpan - http://www.cpan.org/
# Bioperl is available from their site - http://www.bioperl.org/
#
# Last updated 14/06/2004 by Ralf Schmid
# Copyright (C) 2004 Ralf Schmid

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.


use strict;
use File::stat;
use Term::ANSIColor;
use Bio::SeqIO;
use Bio::Tools::pICalculator;
use Bio::Tools::SeqStats;
use DBI;
  use DBD::Pg;  
use Time::localtime;

my $version_number = "0.1            ";
my @PATH=split(":","$ENV{'PATH'}");   


#############################################################################
############################ start main menu ################################
#############################################################################

&options();
#### think about configuration file for following generally used variables
my $seq2phys; # sequences 
my $seq2phys_phys; #tab delimited output file

##################################################################################
#################  1-Calculate physical properties  ##############################
##################################################################################

sub calculate_properties()  {
#### 
  system("clear");
  print colored("\n\t##### Calculate physical properties #####\n","white bold" , "on_black");
  print "\nThis facility uses Bioperl modules to calculate pI, charge at three different"; 
  print "\n pH-values and Mw for a set peptides and stores the results in a PartiGene database\n";
 
#### get sequences, add option for getting sequences from db later
  print "\nPlease enter the name of the file containing the fasta sequences on";
  print "\nwhich you want to perform the calculations. For example ";
  print colored ("yeti.fsa","green bold");  
  print "\nPlease give either the full or relative location of the file.\n";  
  $seq2phys = &get_file();
  $seq2phys_phys = $seq2phys . "_phys"; 


  open (FILEPHYS,">$seq2phys_phys") ||  die "Can't open $seq2phys_phys\n";
  print FILEPHYS "#### id ####\tlen\tundef\tminmw\tmaxmw\tchpos\tchneg\tiep\t c5\t c7\t c9\n";
 
  print "\nCalculating properties now. Please wait ...\n";
 
##### pI to file
  my $in = Bio::SeqIO->new( -file => "$seq2phys",
                            -format => 'Fasta' );
  my $calc = Bio::Tools::pICalculator->new(-places => 2,
                                           -pKset => 'EMBOSS');
  while ( my $seq = $in->next_seq ) {
    my $string = $seq->seq(); 
    my $length = length $string;
    $calc->seq($seq);  
    my $seq_id =  $seq->id;
    my $iep = $calc->iep;              
    my $ch_5 = $calc->charge_at_pH("5") ;
    my $ch_7 = $calc->charge_at_pH("7") ;
    my $ch_9 = $calc->charge_at_pH("9") ;             
    my $seqobj = Bio::PrimarySeq->new(-seq=> "$string",
                                 -alphabet=>'protein',);  
    my $weight = Bio::Tools::SeqStats->get_mol_wt($seqobj);
    my $minmw = $$weight[0];
    my $maxmw = $$weight[1];
    my $hash_ref = Bio::Tools::SeqStats->count_monomers($seqobj);			 
    my %hash = %$hash_ref; # dereference hash to avoid warning
#    my $ch_pos = (%$hash_ref->{"K"} + %$hash_ref->{"R"});
    my $ch_pos = ($hash{"K"} + $hash{"R"});
    my $ch_neg = ($hash{"D"} + $hash{"E"});
    my $undef = '0';
    if ($hash{"X"}) {
       $undef = $hash{"X"};
    }
    print FILEPHYS sprintf ("%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\t%.1f\t%.1f\t%.1f\n",
         $seq_id, $length, $undef, $minmw, $maxmw, $ch_pos, $ch_neg, $iep, $ch_5, $ch_7, $ch_9);
  }
  close FILEPHYS;
  print "\nPhysical properties saved in $seq2phys_phys";
  sleep(5);
}

##################################################################################
#################  2-Databasing  #################################################
##################################################################################
sub databasing()  {
#### parse files, create and populate db
  system("clear");
  print colored("\n\t##### databasing #####\n","white bold" , "on_black");
  print "\nThis menu takes creates and/or populates a PartiGene like\n";
  print "database with the physical properties calculated in option 1\n\n";
 

#### First some checks...pg running, partigene db, step1 output
   
  &postmaster_check;

#### check for partigene.conf file and partigene db
  my $filename = "~/.partigene.conf";
  my $pg_database; my $pg_flag;
  $filename =~ s{ ^ ~ ( [^/]* ) }
              { $1
                    ? (getpwnam($1))[7]
                    : ( $ENV{HOME} || $ENV{LOGDIR}
                         || (getpwuid($>))[7]
                       )
              }ex;

  if  (-e "$filename") {
    open (CONFILE,"$filename") ||  die "Can't open configuration file\n";
    while (my $line=<CONFILE>) {      
      if ($line=~/^DATABASE\=(.+)/i) { $pg_database=$1; }       
    }
    close (CONFILE);
    
#### user has a Partigene db defined in the config file
    print "\nDo you want to use PartiGene database: $pg_database ";
    print "to store physical property information?";
    $pg_flag  = &yes_no();
  }
  
#### don't use Pg database  
  if ($pg_flag != 1) {
    print "What database do you want to create or use?\n";
    $pg_database = <STDIN>; 
    chomp $pg_database; 
  }
  
 
  my $conn=DBI->connect("dbi:Pg:dbname=$pg_database", "", "", {PrintError => 0}); #Last two values would be user/pass.
  my $update_flag = 2;
  if (! $conn)   { ### Couldn't connect to the database  
    print "\nCouldn't connect to the database $pg_database";
    print "\nDo you want to create it?";
    my $answer = &query_for_continue;

    &create_physpropdb($pg_database,"1","1");  ### arguments: <db, does db exist, does table exist>  
  }
  else {
    my @table = $conn->tables('','',undef,'TABLE');
    my $physprop_table_flag=0;        
    for(my $n=0; $n < @table; $n++)    {
      $table[$n] =~ s/public\.//; #get rid of "public." which is present in some versions of DBD.Pg 
      if($table[$n] eq "physprop") { $physprop_table_flag=1; } 
    }
    if ($physprop_table_flag==0) {
      print colored("\n$pg_database does not contain physprop table, do you want to create it?","red bold"); 
      my $answer = &yes_no();
      if ($answer == 1)  {
        &create_physpropdb($pg_database,"0","1");
      }
  
      else {&query_for_exit}
      print "\nphysprop table created.\n";
    }
    else {
      print "\nphysprop entries are already existing for $pg_database.";
      print "\nDo you want to remove old entries? (recommended)";
      print "\n\n- Otherwise you will add new entries to already existing entries,";
      print "\nprevious entries with identical protein identifiers";
      print "\nwill be overwritten";
      my $answer=yes_no();
      if($answer==1)   {#delete existing entries
        print "Deleting old entries. Please wait ...\n";
        my $result = $conn->do("DELETE from physprop;");
        print "Old entries deleted.\n\n";                 
      }
      else {$update_flag = 1;}
            
    }          
    $conn->disconnect or warn "Disconnection failed: $DBI::errstr\n";
  }   
  
#### file check and get physprop results
  my @list = glob("*_phys");
  if ($seq2phys_phys) {#still defined from option 1
    open (FILEPHYS,"<$seq2phys_phys") ||  die "Can't open $seq2phys_phys\n";
  }  
  elsif (@list == 1)  {#assuming the one file in the directory is the one to use
    $seq2phys_phys = $list[0];
    open (FILEPHYS,"<$seq2phys_phys") ||  die "Can't open $seq2phys_phys\n";
  }
  else { 
    print "\nPlease give either the full or relative location of the output file \"somefile_phys\" created in option 1.\n";  
    $seq2phys_phys = &get_file;
  }
  
  open (FILEPHYS,"<$seq2phys_phys") ||  die "Can't open $seq2phys_phys\n";  

  print "\nOutput file $seq2phys_phys found. \nNow parsing results. Please wait ...\n\n";  


  $conn=DBI->connect("dbi:Pg:dbname=$pg_database", "", "", {PrintError => 0}); #Last two values would be user/pass.


  my @data;
  my $n=0;
  if ($update_flag == 1) { ### update    
    while (my $line=<FILEPHYS>) {
      unless ($line=~/^###/) { 
        @data=split("\t",$line);
        my $check = $conn->do("DELETE from physprop WHERE protein_id='$data[0]';");
        my $insert = $conn->do("INSERT INTO physprop (protein_id, num_res, undef_res, minmolw, maxmolw, pos_res, neg_res, pisoel, char_ph5, char_ph7, char_ph9)
	values ('$data[0]','$data[1]','$data[2]','$data[3]','$data[4]','$data[5]','$data[6]','$data[7]','$data[8]','$data[9]','$data[10]');", {PrintError => 0});

        $n++;
        printf("\r%9d sequences processed so far",$n);  
      }
    }
  }
  else { ### just enter new data
    while (my $line=<FILEPHYS>) {
      unless ($line=~/^###/) { 
        @data=split("\t",$line);
        my $insert = $conn->do("INSERT INTO physprop (protein_id, num_res, undef_res, minmolw, maxmolw, pos_res, neg_res, pisoel, char_ph5, char_ph7, char_ph9)
	values ('$data[0]','$data[1]','$data[2]','$data[3]','$data[4]','$data[5]','$data[6]','$data[7]','$data[8]','$data[9]','$data[10]');", {PrintError => 0});

        $n++;
        printf("\r%9d sequences processed so far",$n);  
      }
    }
  } 
  $conn->disconnect or warn "Disconnection failed: $DBI::errstr\n"; 
  &query_for_exit();     
}




###################################################################################
###################################################################################
###                                                                             ###
###                               all the subs                                  ###
###                                                                             ###
###################################################################################
###################################################################################

#########################################################################################################################
sub options () {
#### option selection
   my $answer=0;
   while($answer!=3) {
      $answer=&title_page();
      if($answer==1)  { &calculate_properties(); } 
      if($answer==2)  { &databasing(); } 
     
   }
   system("clear");
   exit();   
}
############################################################################################################################

   
###################################################################################
sub title_page() {
#### intro & sub-menu selection
    print_title();
    print "\n\tBefore proceeding please make sure you are in an appropriate\n";
    print "\tdirectory.\n\n";
    print "\t\t1. Calculate physical properties\n";
    print "\t\t2. Databasing\n";
    print "\t\t3. Quit.\n";

    my $flag=0; my $answer;
    while($flag==0) {
        $answer=<>;
	if($answer=~/^[1|2|3]$/) { $flag=1; next; }
        else {print " You have entered $answer This is not an option. Please try again\n";}
    }
    return $answer;
}
####################################################################################


############################################################################################################################
sub print_title() {
##### displays title
    print colored("\n\n\n", "white bold", "on_black"); 
    print colored("\t###########################################################\n","white bold", "on_black");
    print colored("\t###                                                     ###\n","white bold", "on_black");
    print colored("\t###             annot8r_physprop.pl                     ###\n","white bold", "on_black");
    print colored("\t###     a script to calculate physical properties       ###\n","white bold", "on_black");
    print colored("\t###     for peptide sequences Vs $version_number        ###\n","white bold", "on_black");
    print colored("\t###                                                     ###\n","white bold", "on_black");
    print colored("\t###     EGTDC 2004                                      ###\n","white bold", "on_black");
    print colored("\t###                                                     ###\n","white bold", "on_black");
    print colored("\t###     For news and upgrades and help:                 ###\n","white bold", "on_black");
    print colored("\t###     nematodes.bioinf\@ed.ac.uk                       ###\n","white bold", "on_black");
    print colored("\t###                                                     ###\n","white bold", "on_black");
    print colored("\t###     Help for EG-Awardees:                           ###\n","white bold", "on_black");
    print colored("\t###     helpdesk\@envgen.nox.ac.uk                       ###\n","white bold", "on_black");
    print colored("\t###                                                     ###\n","white bold", "on_black");
    print colored("\t###########################################################\n\n\n","white bold", "on_black");
   
}
###########################################################################################################################


###################################################################################
sub get_file() {    
#### get file from user input, check existence
  my $flag = 0; my $file_name;
  while ($flag == 0) {
    $file_name = <STDIN>;
    chomp $file_name;
    $file_name =~ s/\s+//g;
    if (-e $file_name) {$flag = 1;}
    else {
      print "\nCan't find $file_name, do you want to try again?";
      my $answer = &yes_no();
      if ($answer != 1) {exit();} 
    }  
  }
  return "$file_name";
}  
###################################################################################


####################################################################################
sub yes_no() {
#### returns 1 for y
  my $yflag=0;
  print colored(" [y/n] : ","green bold");
  my $input='';
  while($input!~/y|n/i)   {
    print "\b";
    $input=<STDIN>;
    chomp $input;
  }
  if($input=~/^y/i) { $yflag=1; }
  return $yflag;
}
####################################################################################


 #######################################################################################
sub find_program()  {
#### search for argument in path, exit if not found
  my $prog=$_[0];
  my $pathflag=0;
  my $path;
  my $finalpath;
  foreach $path (@PATH) {
    if (-f "$path/$prog") {
      $pathflag=1; $finalpath=$path; last;
    }
  }
  if($pathflag==0)   { 
    print colored("\nCan't find the $prog utility on your system\n","red bold");
    print colored("Please ensure it is installed and in your path\n\n","red bold");
    exit();
  }
  else  {
    return "$finalpath/$prog";
  }
}
###############################################################################################################

#########################################################################################################################
sub postmaster_check() {
#### check for postmaster/postgresql process
  my $postmaster=`ps -e|grep postmaster`; ### See if the process is running

  if(!$postmaster)  {
    print colored("\n#### Postmaster is not running ####\n","red bold");
    print colored("Please ensure that postgreSQL is correctly installed and running\n","red bold");
    exit();
  }
  else {print "CHECK POSTGRESQL   OK => Postmaster running\n\n"} 

#### check whether user does exist   
  my $user_status = system ("psql -l > /dev/null"); ### command will fail unless (postgresql) user exists      
  unless ($user_status == 0) {
    my $username = `whoami`; chomp $username;
    print colored("\n\t#### CONNECTION TO POSTGRESQL FAILED ####\n","red bold");
    print colored("Most likely you have forgotten to run \"createuser $username\"\n","red bold");
    print colored("during the postgreSQL setup\n","red bold");
    exit();
  }
}
######################################################################################################################


###########################################################################################################################
sub query_for_exit() {
#exits program if 'n' entered back to main for y
   print "\nWould you like to continue? ";
   print colored(" [y/n] : ","green bold");
   my $input='';
   while($input!~/y|n/i)   {
     print "\b";
     $input=<>;
     chomp $input;
  }
  if ($input=~/^y/i) { print "Back to main menu\n"; &options;}
  if ($input=~/^n/i) { print "Exiting the program\n"; exit(); }
}
####################################################################################


###########################################################################################################################
sub query_for_continue() {
#exits program if 'n' entered carries on for 'y'
#   print "\nWould you like to continue? ";
   print colored(" [y/n] : ","green bold");
   my $input='';
   while($input!~/y|n/i)   {
     print "\b";
     $input=<>;
     chomp $input;
  }
  if ($input=~/^n/i) { print "Exiting the program\n"; exit(); }
}
####################################################################################


##############################################################################################
sub create_physpropdb()  { 
#### one basic table   
  my $database=shift;
  my $new_db=shift;
  my $table_flag = shift;
  my $createdb_exe= &find_program("createdb");
  if($new_db == 1) { system("$createdb_exe $database >& /dev/null"); }  
  my $conn=DBI->connect("dbi:Pg:dbname=$database", "", "");
  if($table_flag == 1) {
    my $result=$conn->do ("create table physprop (protein_id varchar(20) not null primary key, num_res int, undef_res int, minmolw int, 
    maxmolw int, pos_res int, neg_res int, pisoel decimal(5,2),  char_ph5 decimal(8,2), char_ph7 decimal(8,2), char_ph9 decimal(8,2));");  
  }
  my $errorMessage = $conn->errstr;
  if ($errorMessage) {print "$errorMessage\n";}
  $conn->disconnect or warn "Disconnection failed: $DBI::errstr\n";
}

###############################################################################################################
###   fin de siecle
###############################################################################################################
